@@ -8,42 +8,41 @@
  *)
 
 open Lwt
-open Lwt_react
 open LTerm_geom
 open LTerm_text
 open LTerm_key
 
 let rec loop ui coord =
   LTerm_ui.wait ui >>= function
-    | LTerm_event.Key{ code = Up } ->
+    | LTerm_event.Key{ code = Up; _ } ->
         coord := { !coord with row = !coord.row - 1 };
         LTerm_ui.draw ui;
         loop ui coord
-    | LTerm_event.Key{ code = Down } ->
+    | LTerm_event.Key{ code = Down; _ } ->
         coord := { !coord with row = !coord.row + 1 };
         LTerm_ui.draw ui;
         loop ui coord
-    | LTerm_event.Key{ code = Left } ->
+    | LTerm_event.Key{ code = Left; _ } ->
         coord := { !coord with col = !coord.col - 1 };
         LTerm_ui.draw ui;
         loop ui coord
-    | LTerm_event.Key{ code = Right } ->
+    | LTerm_event.Key{ code = Right; _ } ->
         coord := { !coord with col = !coord.col + 1 };
         LTerm_ui.draw ui;
         loop ui coord
-    | LTerm_event.Key{ code = Escape } ->
+    | LTerm_event.Key{ code = Escape; _ } ->
         return ()
-    | ev ->
+    | _ ->
         loop ui coord
 
 let draw ui matrix coord =
   let size = LTerm_ui.size ui in
   let ctx = LTerm_draw.context matrix size in
   LTerm_draw.clear ctx;
-  LTerm_draw.draw_frame ctx { row1 = 0; col1 = 0; row2 = size.rows; col2 = size.cols } LTerm_draw.Light;
+  LTerm_draw.draw_frame_labelled ctx { row1 = 0; col1 = 0; row2 = size.rows; col2 = size.cols } ~alignment:H_align_center (Zed_string.of_utf8 "Use arrow keys to move text 文字") LTerm_draw.Light;
   if size.rows > 2 && size.cols > 2 then begin
     let ctx = LTerm_draw.sub ctx { row1 = 1; col1 = 1; row2 = size.rows - 1; col2 = size.cols - 1 } in
-    LTerm_draw.draw_styled ctx coord.row coord.col (eval [B_fg LTerm_style.lblue; S"Move me"; E_fg])
+    LTerm_draw.draw_styled ctx coord.row coord.col (eval [B_fg LTerm_style.lblue; S "Move m̀é 囧"; E_fg])
   end
 
 let main () =
