@@ -7,7 +7,9 @@
  * This file is a part of Lambda-Term.
  *)
 
-open CamomileLibraryDyn.Camomile
+let pervasives_compare= compare
+
+open CamomileLibraryDefault.Camomile
 open Zed_edit
 open LTerm_key
 open LTerm_geom
@@ -49,8 +51,8 @@ let actions = [
   Add_macro_counter, "add-macro-counter";
 ]
 
-let actions_to_names = Array.of_list (List.sort (fun (a1, n1) (a2, n2) -> Pervasives.compare a1 a2) actions)
-let names_to_actions = Array.of_list (List.sort (fun (a1, n1) (a2, n2) -> Pervasives.compare n1 n2) actions)
+let actions_to_names = Array.of_list (List.sort (fun (a1, _) (a2, _) -> pervasives_compare a1 a2) actions)
+let names_to_actions = Array.of_list (List.sort (fun (_, n1) (_, n2) -> pervasives_compare n1 n2) actions)
 
 let action_of_name x =
   let rec loop a b =
@@ -59,7 +61,7 @@ let action_of_name x =
     else
       let c = (a + b) / 2 in
       let action, name = Array.unsafe_get names_to_actions c in
-      match Pervasives.compare x name with
+      match pervasives_compare x name with
         | d when d < 0 ->
             loop a c
         | d when d > 0 ->
@@ -76,7 +78,7 @@ let name_of_action x =
     else
       let c = (a + b) / 2 in
       let action, name = Array.unsafe_get actions_to_names c in
-      match Pervasives.compare x action with
+      match pervasives_compare x action with
         | d when d < 0 ->
             loop a c
         | d when d > 0 ->
@@ -157,22 +159,30 @@ let macro = Zed_macro.create []
 let regexp_word =
   let set = UCharInfo.load_property_set `Alphabetic in
   let set = List.fold_left (fun set ch -> USet.add (UChar.of_char ch) set) set ['0'; '1'; '2'; '3'; '4'; '5'; '6'; '7'; '8'; '9'] in
-  Zed_re.compile (`Repn(`Set set, 1, None))
+  Zed_re.Core.compile (`Repn(`Set set, 1, None))
 
 let dummy_engine = Zed_edit.create ()
 let dummy_cursor = Zed_edit.new_cursor dummy_engine
 let dummy_context = Zed_edit.context dummy_engine dummy_cursor
-let newline = UChar.of_char '\n'
+let newline = Zed_char.unsafe_of_uChar (UChar.of_char '\n')
 
-class edit ?(clipboard = clipboard) ?(macro = macro) () =
+class scrollable = object
+  inherit LTerm_widget.scrollable
+  method! calculate_range page_size document_size = (document_size - page_size/2)
+end
+
+class edit ?(clipboard = clipboard) ?(macro = macro) ?(size = { cols = 1; rows = 1 }) () =
   let locale, set_locale = S.create None in
 object(self)
-  inherit LTerm_widget.t "edit"
+  inherit LTerm_widget.t "edit" as super
+
+  val vscroll = new scrollable
+  method vscroll = vscroll
 
   method clipboard = clipboard
   method macro = macro
 
-  method can_focus = true
+  method! can_focus = true
 
   val mutable engine = dummy_engine
   method engine = engine
@@ -188,14 +198,14 @@ object(self)
   val mutable style = LTerm_style.none
   val mutable marked_style = LTerm_style.none
   val mutable current_line_style = LTerm_style.none
-  method update_resources =
+  method! update_resources =
     let rc = self#resource_class and resources = self#resources in
     style <- LTerm_resources.get_style rc resources;
     marked_style <- LTerm_resources.get_style (rc ^ ".marked") resources;
     current_line_style <- LTerm_resources.get_style (rc ^ ".current-line") resources
 
-  method editable pos len = true
-  method match_word text pos = match_by_regexp regexp_word text pos
+  method editable _pos _len = true
+  method match_word text pos = match_by_regexp_core regexp_word text pos
   method locale = S.value locale
   method set_locale locale = set_locale locale
 
@@ -205,6 +215,50 @@ object(self)
   val mutable local_bindings = Bindings.empty
   method bind keys actions = local_bindings <- Bindings.add keys actions local_bindings
 
+  val mutable shift_width = 0
+  val mutable start = 0
+  val mutable start_line = 0
+  val mutable size = size
+
+  method! size_request = size
+
+  method private update_window_position =
+    let line_set = Zed_edit.lines engine in
+    let line_count = Zed_lines.count line_set in
+    let cursor_offset = Zed_cursor.get_position cursor in
+    let cursor_line = Zed_lines.line_index line_set cursor_offset in
+    let cursor_column = cursor_offset - Zed_lines.line_start line_set cursor_line in
+    let column_display= Zed_lines.force_width line_set (Zed_lines.line_start line_set cursor_line) cursor_column in
+
+
+    (*** check cursor position is in view *)
+
+    (* Horizontal check *)
+    if column_display < shift_width || column_display >= shift_width + size.cols then begin
+
+      shift_width <- max 0 (column_display - size.cols / 2);
+    end;
+
+    (* Vertical check *)
+    let start_line' = Zed_lines.line_index line_set start in
+    let start_line' =
+      if cursor_line < start_line' || cursor_line >= start_line' + size.rows then begin
+        (*let start_line' = max 0 (cursor_line - size.rows / 2) in*)
+        let line_count = Zed_lines.count line_set in
+        let start_line' = min line_count (max 0 (cursor_line - size.rows / 2)) in
+        start <- Zed_lines.line_start line_set start_line';
+        start_line'
+      end else
+        start_line'
+    in
+    (* document size *)
+    if start_line <> start_line' then begin
+      start_line <- start_line';
+      vscroll#set_offset ~trigger_callback:false start_line
+    end;
+    vscroll#set_document_size (line_count+1);
+    ()
+
   initializer
     engine <- (
       Zed_edit.create
@@ -217,7 +271,9 @@ object(self)
     cursor <- Zed_edit.new_cursor engine;
     context <- Zed_edit.context engine cursor;
     Zed_edit.set_data engine (self :> edit);
-    event <- E.map (fun _ -> self#queue_draw) (Zed_edit.update engine [cursor]);
+    event <- E.map (fun _ ->
+      self#update_window_position;
+      self#queue_draw) (Zed_edit.update engine [cursor]);
     self#on_event
       (function
          | LTerm_event.Key key -> begin
@@ -254,7 +310,7 @@ object(self)
                          exec (Zed_macro.contents macro @ actions)
                      | Insert_macro_counter :: actions ->
                          Zed_macro.add macro Insert_macro_counter;
-                         Zed_edit.insert context (Zed_rope.of_string (string_of_int (Zed_macro.get_counter macro)));
+                         Zed_edit.insert context (Zed_rope.of_string (Zed_string.unsafe_of_utf8 (string_of_int (Zed_macro.get_counter macro))));
                          Zed_macro.add_counter macro 1;
                          exec actions
                      | (Add_macro_counter | Set_macro_counter) :: actions ->
@@ -270,7 +326,7 @@ object(self)
                    if resolver = None then
                      match key with
                        | { control = false; meta = false; shift = false; code = Char ch } ->
-                           Zed_edit.insert context (Zed_rope.singleton ch);
+                           Zed_edit.insert_char context ch;
                            true
                        | _ ->
                            false
@@ -282,40 +338,51 @@ object(self)
          | _ ->
              false)
 
-  val mutable shift = 0
-  val mutable start = 0
 
-  method draw ctx focused =
+  method! set_allocation rect =
+    size <- size_of_rect rect;
+    super#set_allocation rect;
+    vscroll#set_page_size size.rows;
+    start <- 0; shift_width <- 0; start_line <- 0;
+    self#update_window_position
+
+  initializer vscroll#on_offset_change (fun n ->
+
+    (* find what line the cursor is currently on. *)
+    let line_set = Zed_edit.lines engine in
+    let cursor_offset = Zed_cursor.get_position cursor in
+    let cursor_line = Zed_lines.line_index line_set cursor_offset in
+
+    start_line <- n;
+    start <- Zed_lines.line_start line_set start_line;
+
+    if cursor_line < start_line then begin
+      let d = start_line - cursor_line in
+      Zed_edit.move_line context d (* first row *)
+    end else if cursor_line >= start_line + size.rows then begin
+      let line_count = Zed_lines.count line_set in
+      let line = max 0 (min (line_count+1) (start_line + size.rows - 1)) in (* last row *)
+      let d = line - cursor_line in
+      Zed_edit.move_line context d
+    end;
+    self#queue_draw;
+  )
+
+
+  method! draw ctx _focused =
     let open LTerm_draw in
 
     let size = LTerm_draw.size ctx in
 
-    (*** Check that the cursor is displayed ***)
-
     let line_set = Zed_edit.lines engine in
     let cursor_offset = Zed_cursor.get_position cursor in
     let cursor_line = Zed_lines.line_index line_set cursor_offset in
     let cursor_column = cursor_offset - Zed_lines.line_start line_set cursor_line in
 
-    (* Horizontal check *)
-    if cursor_column < shift || cursor_column >= shift + size.cols then
-      shift <- max 0 (cursor_column - size.cols / 2);
-
-    (* Vertical check *)
-    let start_line = Zed_lines.line_index line_set start in
-    let start_line =
-      if cursor_line < start_line || cursor_line >= start_line + size.rows then begin
-        let start_line = max 0 (cursor_line - size.rows / 2) in
-        start <- Zed_lines.line_start line_set start_line;
-        start_line
-      end else
-        start_line
-    in
-
     (*** Drawing ***)
 
     (* Initialises points with the text style and spaces. *)
-    fill ctx (UChar.of_char ' ');
+    fill ctx (Zed_char.unsafe_of_char ' ');
     fill_style ctx style;
 
     (*** Text drawing ***)
@@ -334,7 +401,7 @@ object(self)
             if row < size.rows then skip_eol row zip
           end else begin
             draw_char ctx row col char;
-            draw_line row (col + 1) zip
+            draw_line row (col + (Zed_char.width char)) zip
           end
         end
 
@@ -349,8 +416,8 @@ object(self)
           skip_eol row zip
 
     and skip_bol row zip remaining =
-      if remaining = 0 then
-        draw_line row 0 zip
+      if remaining <= 0 then
+        draw_line row (-remaining) zip
       else if Zed_rope.Zip.at_eos zip then
         draw_eoi (row + 1)
       else
@@ -359,17 +426,17 @@ object(self)
           let row = row + 1 in
           if row < size.rows then begin_line row zip
         end else
-          skip_bol row zip (remaining - 1)
+          skip_bol row zip (remaining - (Zed_char.width char))
 
     and begin_line row zip =
       if Zed_rope.Zip.at_eos zip then
         draw_eoi row
-      else if shift <> 0 then begin
-        skip_bol row zip shift
+      else if shift_width <> 0 then begin
+        skip_bol row zip shift_width
       end else
         draw_line row 0 zip
 
-    and draw_eoi row =
+    and draw_eoi _row =
       ()
     in
 
@@ -428,11 +495,13 @@ object(self)
       end
     end
 
-  method cursor_position =
+  method! cursor_position =
     let line_set = Zed_edit.lines engine in
     let cursor_offset = Zed_cursor.get_position cursor in
     let cursor_line = Zed_lines.line_index line_set cursor_offset in
-    let cursor_column = cursor_offset - Zed_lines.line_start line_set cursor_line in
+    let line_start= Zed_lines.line_start line_set cursor_line  in
     let start_line = Zed_lines.line_index line_set start in
-    Some { row = cursor_line - start_line; col = cursor_column - shift }
+    let col= Zed_lines.force_width line_set line_start (cursor_offset - line_start)
+        - shift_width in
+    Some { row = cursor_line - start_line; col }
 end
