@@ -7,18 +7,19 @@
  * This file is a part of Lambda-Term.
  *)
 
-open CamomileLibraryDyn.Camomile
+let pervasives_compare= compare
+
+open CamomileLibraryDefault.Camomile
 open Lwt_react
 open LTerm_geom
 open LTerm_style
-open LTerm_text
 open LTerm_key
 
 let return, (>>=) = Lwt.return, Lwt.(>>=)
 
 exception Interrupt
 type prompt = LTerm_text.t
-type history = Zed_utf8.t list
+type history = Zed_string.t list
 
 (* +-----------------------------------------------------------------+
    | Completion                                                      |
@@ -42,8 +43,26 @@ let common_prefix = function
   | [] -> ""
   | word :: rest -> List.fold_left common_prefix_one word rest
 
+let zed_common_prefix_one a b =
+  let rec loop ofs =
+    if ofs = Zed_string.bytes a || ofs = Zed_string.bytes b then
+      Zed_string.sub_ofs ~ofs:0 ~len:ofs a
+    else
+      let ch1, ofs1= Zed_string.extract_next a ofs
+      and ch2, ofs2= Zed_string.extract_next b ofs in
+      if ch1 = ch2 && ofs1 = ofs2 then
+        loop ofs1
+      else
+        Zed_string.sub_ofs ~ofs:0 ~len:ofs a
+  in
+  loop 0
+
+let zed_common_prefix = function
+  | [] -> Zed_string.empty ()
+  | word :: rest -> List.fold_left zed_common_prefix_one word rest
+
 let lookup word words = List.filter (fun word' -> Zed_utf8.starts_with word' word) words
-let lookup_assoc word words = List.filter (fun (word', x) -> Zed_utf8.starts_with word' word) words
+let lookup_assoc word words = List.filter (fun (word', _) -> Zed_utf8.starts_with word' word) words
 
 (* +-----------------------------------------------------------------+
    | Actions                                                         |
@@ -60,6 +79,8 @@ type action =
   | Complete_bar
   | History_prev
   | History_next
+  | History_search_prev
+  | History_search_next
   | Accept
   | Clear_screen
   | Prev_search
@@ -80,6 +101,8 @@ let doc_of_action = function
   | Complete_bar -> "complete current input using the completion bar."
   | History_prev -> "go to the previous entry of the history."
   | History_next -> "go to the next entry of the history."
+  | History_search_prev -> "go to the previous entry of the history that matches the start of the current line."
+  | History_search_next -> "go to the next entry of the history that matches the start of the current line."
   | Accept -> "accept the current input."
   | Clear_screen -> "clear the screen."
   | Prev_search -> "search backward in the history."
@@ -99,6 +122,8 @@ let actions = [
   Complete_bar, "complete-bar";
   History_prev, "history-prev";
   History_next, "history-next";
+  History_search_prev, "history-search-prev";
+  History_search_next, "history-search-next";
   Accept, "accept";
   Clear_screen, "clear-screen";
   Prev_search, "prev-search";
@@ -109,8 +134,8 @@ let actions = [
   Edit_with_external_editor, "edit-with-external-editor";
 ]
 
-let actions_to_names = Array.of_list (List.sort (fun (a1, n1) (a2, n2) -> Pervasives.compare a1 a2) actions)
-let names_to_actions = Array.of_list (List.sort (fun (a1, n1) (a2, n2) -> Pervasives.compare n1 n2) actions)
+let actions_to_names = Array.of_list (List.sort (fun (a1, _) (a2, _) -> pervasives_compare a1 a2) actions)
+let names_to_actions = Array.of_list (List.sort (fun (_, n1) (_, n2) -> pervasives_compare n1 n2) actions)
 
 let action_of_name x =
   let rec loop a b =
@@ -119,7 +144,7 @@ let action_of_name x =
     else
       let c = (a + b) / 2 in
       let action, name = Array.unsafe_get names_to_actions c in
-      match Pervasives.compare x name with
+      match pervasives_compare x name with
         | d when d < 0 ->
             loop a c
         | d when d > 0 ->
@@ -136,7 +161,7 @@ let name_of_action x =
     else
       let c = (a + b) / 2 in
       let action, name = Array.unsafe_get actions_to_names c in
-      match Pervasives.compare x action with
+      match pervasives_compare x action with
         | d when d < 0 ->
             loop a c
         | d when d > 0 ->
@@ -190,7 +215,29 @@ let () =
    | The read-line engine                                            |
    +-----------------------------------------------------------------+ *)
 
+let is_prefix ~prefix s=
+  let prefix= Zed_string.to_utf8 prefix
+  and s= Zed_string.to_utf8 s in
+  String.length prefix <= String.length s &&
+  (let i = ref 0 in
+   while !i < String.length prefix && s.[!i] = prefix.[!i] do incr i done;
+   !i = String.length prefix
+  )
+
+let history_find predicate history =
+  let rec history_find_ skipped = function
+    | [] -> None
+    | x :: xs ->
+      if predicate x then
+        Some (skipped, x, xs)
+      else
+        history_find_ (x :: skipped) xs
+  in
+  history_find_ [] history
+
 let search_string str sub =
+  let str= Zed_string.to_utf8 str
+  and sub= Zed_string.to_utf8 sub in
   let rec equal_at a b =
     (b = String.length sub) || (String.unsafe_get str a = String.unsafe_get sub b) && equal_at (a + 1) (b + 1)
   in
@@ -217,7 +264,7 @@ type completion_state = {
   start : int; (* Beginning of the word being completed *)
   index : int; (* Index of the selected in [words]      *)
   count : int; (* Length of [words]                     *)
-  words : (Zed_utf8.t * Zed_utf8.t) list;
+  words : (Zed_string.t * Zed_string.t) list;
 }
 
 let no_completion = {
@@ -230,9 +277,9 @@ let no_completion = {
 type direction = Forward | Backward
 
 type search_status =
-  { before : Zed_utf8.t list
-  ; after  : Zed_utf8.t list
-  ; match_ : (Zed_utf8.t * int) option
+  { before : Zed_string.t list
+  ; after  : Zed_string.t list
+  ; match_ : (Zed_string.t * int) option
   }
 
 class virtual ['a] engine ?(history = []) ?(clipboard = LTerm_edit.clipboard) ?(macro = macro) () =
@@ -255,6 +302,13 @@ class virtual ['a] engine ?(history = []) ?(clipboard = LTerm_edit.clipboard) ?(
   let completion_index = S.map          (fun c -> c.index) completion_state in
   let history, set_history = S.create (history, []) in
   let message, set_message = S.create None in
+  let history_prefix, set_history_prefix =
+    let ev, send = E.create () in
+    let edit_changes = Zed_edit.changes edit in
+    let edit_changes = E.map (fun _ -> Zed_edit.text edit) edit_changes in
+    let prefix = S.hold (Zed_rope.empty ()) (E.select [ev; edit_changes]) in
+    prefix, send
+  in
 object(self)
   method virtual eval : 'a
   method edit = edit
@@ -270,7 +324,7 @@ object(self)
   val mutable completion_event = E.never
 
   (* Save for when setting the macro counter. *)
-  val mutable save = (0, Zed_rope.empty)
+  val mutable save = (0, Zed_rope.empty ())
 
   method set_completion ?(index=0) start words =
     let count = List.length words in
@@ -306,14 +360,20 @@ object(self)
     let comp = S.value completion_state in
     let prefix_length = Zed_edit.position context - comp.start in
     match comp.words with
-      | [] ->
-          ()
-      | [(completion, suffix)] ->
-          Zed_edit.insert context (Zed_rope.of_string (Zed_utf8.after completion prefix_length));
-          Zed_edit.insert context (Zed_rope.of_string suffix)
-      | (completion, suffix) :: rest ->
-          let word = List.fold_left (fun acc (word, _) -> common_prefix_one acc word) completion rest in
-          Zed_edit.insert context (Zed_rope.of_string (Zed_utf8.after word prefix_length))
+    | [] -> ()
+    | [(completion, suffix)] ->
+      Zed_edit.insert context (Zed_rope.of_string
+        (Zed_string.sub completion ~pos:prefix_length
+          ~len:(Zed_string.length completion - prefix_length)));
+      Zed_edit.insert context (Zed_rope.of_string suffix)
+    | (completion, _suffix) :: rest ->
+      let word = List.fold_left
+        (fun acc (word, _) -> zed_common_prefix_one acc word)
+        completion rest
+      in
+      Zed_edit.insert context (Zed_rope.of_string
+        (Zed_string.sub word ~pos:prefix_length
+          ~len:(Zed_string.length word - prefix_length)))
 
   (* The event which search for the string in the history. *)
   val mutable search_event = E.never
@@ -345,12 +405,12 @@ object(self)
         match entries with
         | [] ->
           set_status other_entries entries None;
-          set_message (Some(LTerm_text.of_string "Reverse search: not found"))
+          set_message (Some(LTerm_text.of_utf8 "Reverse search: not found"))
         | entry :: rest ->
           match search_string entry input with
           | Some pos -> begin
               match search_status with
-              | Some { match_ = Some (entry', _) } when entry = entry' ->
+              | Some { match_ = Some (entry', _); _ } when entry = entry' ->
                 loop (entry :: other_entries) rest
               | _ ->
                 set_status other_entries rest (Some (entry, pos));
@@ -360,7 +420,7 @@ object(self)
                   txt.(i) <- (ch, { style with underline = Some true })
                 done;
                 set_message
-                  (Some (Array.append (LTerm_text.of_string "Reverse search: ") txt))
+                  (Some (Array.append (LTerm_text.of_utf8 "Reverse search: ") txt))
             end
           | None ->
             loop (entry :: other_entries) rest
@@ -400,7 +460,7 @@ object(self)
       ()
 
   method insert ch =
-    Zed_edit.insert context (Zed_rope.singleton ch)
+    Zed_edit.insert_char context ch
 
   method send_action action =
     if action <> Edit LTerm_edit.Stop_macro then Zed_macro.add macro action;
@@ -409,12 +469,12 @@ object(self)
           set_mode Edition;
           set_message None;
           match search_status with
-            | Some { match_ = Some (entry, pos) } ->
+            | Some { match_ = Some (entry, _pos); _ } ->
                 search_status <- None;
                 Zed_edit.goto context 0;
                 Zed_edit.remove context (Zed_rope.length (Zed_edit.text edit));
                 Zed_edit.insert context (Zed_rope.of_string entry)
-            | Some { match_ = None } | None ->
+            | Some { match_ = None; _ } | None ->
                 ()
         end
 
@@ -455,7 +515,8 @@ object(self)
           if comp.words <> [] then begin
             let prefix_length = Zed_edit.position context - comp.start in
             let completion, suffix = List.nth comp.words comp.index in
-            Zed_edit.insert context (Zed_rope.of_string (Zed_utf8.after completion prefix_length));
+            Zed_edit.insert context (Zed_rope.of_string
+              (Zed_string.after completion prefix_length));
             Zed_edit.insert context (Zed_rope.of_string suffix)
           end
 
@@ -485,6 +546,36 @@ object(self)
                 Zed_edit.insert context (Zed_rope.of_string line)
         end
 
+      | History_search_prev when S.value mode = Edition -> begin
+          let prev, next = S.value history in
+          let text = Zed_rope.to_string @@ Zed_edit.text edit in
+          let prefix = S.value history_prefix in
+          match history_find (is_prefix ~prefix:(Zed_rope.to_string prefix)) prev with
+          | None ->
+            ()
+          | Some (not_matched, line, rest) ->
+            set_history (rest, not_matched @ text :: next);
+            Zed_edit.goto context 0;
+            Zed_edit.delete_next_line context;
+            Zed_edit.insert context (Zed_rope.of_string line);
+            set_history_prefix prefix
+        end
+
+      | History_search_next when S.value mode = Edition -> begin
+          let prev, next = S.value history in
+          let prefix = S.value history_prefix in
+          match history_find (is_prefix ~prefix:(Zed_rope.to_string prefix)) next with
+          | None ->
+            ()
+          | Some (not_matched, line, rest) ->
+            let text = Zed_rope.to_string @@ Zed_edit.text edit in
+            set_history (not_matched @ text :: prev, rest);
+            Zed_edit.goto context 0;
+            Zed_edit.delete_next_line context;
+            Zed_edit.insert context (Zed_rope.of_string line);
+            set_history_prefix prefix
+        end
+
       | Prev_search -> self#search Backward
       | Next_search -> self#search Forward
 
@@ -508,7 +599,7 @@ object(self)
           List.iter self#send_action (Zed_macro.contents macro)
 
       | Edit LTerm_edit.Insert_macro_counter ->
-          Zed_edit.insert context (Zed_rope.of_string (string_of_int (Zed_macro.get_counter macro)));
+          Zed_edit.insert context (Zed_rope.of_string (Zed_string.unsafe_of_utf8 (string_of_int (Zed_macro.get_counter macro))));
           Zed_macro.add_counter macro 1
 
       | Edit LTerm_edit.Set_macro_counter when S.value mode = Edition ->
@@ -517,7 +608,7 @@ object(self)
           Zed_edit.goto context 0;
           Zed_edit.remove context (Zed_rope.length text);
           set_mode Set_counter;
-          set_message (Some (LTerm_text.of_string "Enter a value for the macro counter."))
+          set_message (Some (LTerm_text.of_utf8 "Enter a value for the macro counter."))
 
       | Edit LTerm_edit.Add_macro_counter when S.value mode = Edition ->
           let text = Zed_edit.text edit in
@@ -525,7 +616,7 @@ object(self)
           Zed_edit.goto context 0;
           Zed_edit.remove context (Zed_rope.length text);
           set_mode Add_counter;
-          set_message (Some (LTerm_text.of_string "Enter a value to add to the macro counter."))
+          set_message (Some (LTerm_text.of_utf8 "Enter a value to add to the macro counter."))
 
       | Accept -> begin
           match S.value mode with
@@ -533,9 +624,9 @@ object(self)
                 ()
             | Set_counter ->
                 let pos, text = save in
-                save <- (0, Zed_rope.empty);
+                save <- (0, Zed_rope.empty ());
                 (try
-                   Zed_macro.set_counter macro (int_of_string (Zed_rope.to_string (Zed_edit.text edit)))
+                   Zed_macro.set_counter macro (int_of_string (Zed_string.to_utf8 (Zed_rope.to_string (Zed_edit.text edit))))
                  with Failure _ ->
                    ());
                 Zed_edit.goto context 0;
@@ -546,9 +637,9 @@ object(self)
                 set_message None
             | Add_counter ->
                 let pos, text = save in
-                save <- (0, Zed_rope.empty);
+                save <- (0, Zed_rope.empty ());
                 (try
-                   Zed_macro.add_counter macro (int_of_string (Zed_rope.to_string (Zed_edit.text edit)))
+                   Zed_macro.add_counter macro (int_of_string (Zed_string.to_utf8 (Zed_rope.to_string (Zed_edit.text edit))))
                  with Failure _ ->
                    ());
                 Zed_edit.goto context 0;
@@ -591,13 +682,13 @@ class virtual ['a] abstract = object
   method virtual clipboard : Zed_edit.clipboard
   method virtual macro : action Zed_macro.t
   method virtual stylise : bool -> LTerm_text.t * int
-  method virtual history : (Zed_utf8.t list * Zed_utf8.t list) signal
+  method virtual history : (Zed_string.t list * Zed_string.t list) signal
   method virtual message : LTerm_text.t option signal
   method virtual input_prev : Zed_rope.t
   method virtual input_next : Zed_rope.t
-  method virtual completion_words : (Zed_utf8.t * Zed_utf8.t) list signal
+  method virtual completion_words : (Zed_string.t * Zed_string.t) list signal
   method virtual completion_index : int signal
-  method virtual set_completion : ?index:int -> int -> (Zed_utf8.t * Zed_utf8.t) list -> unit
+  method virtual set_completion : ?index:int -> int -> (Zed_string.t * Zed_string.t) list -> unit
   method virtual completion : unit
   method virtual complete : unit
   method virtual show_box : bool
@@ -609,33 +700,33 @@ end
    +-----------------------------------------------------------------+ *)
 
 class read_line ?history () = object(self)
-  inherit [Zed_utf8.t] engine ?history ()
+  inherit [Zed_string.t] engine ?history ()
   method eval = Zed_rope.to_string (Zed_edit.text self#edit)
 end
 
 class read_password () = object(self)
-  inherit [Zed_utf8.t] engine () as super
+  inherit [Zed_string.t] engine () as super
 
-  method stylise last =
+  method! stylise last =
     let text, pos = super#stylise last in
     for i = 0 to Array.length text - 1 do
-      let ch, style = text.(i) in
-      text.(i) <- (UChar.of_char '*', style)
+      let _ch, style = text.(i) in
+      text.(i) <- (Zed_char.unsafe_of_char '*', style)
     done;
     (text, pos)
 
   method eval = Zed_rope.to_string (Zed_edit.text self#edit)
 
-  method show_box = false
+  method! show_box = false
 
-  method send_action = function
+  method! send_action = function
     | Prev_search | Next_search -> ()
     | action -> super#send_action action
 end
 
 type 'a read_keyword_result =
   | Rk_value of 'a
-  | Rk_error of Zed_utf8.t
+  | Rk_error of Zed_string.t
 
 class ['a] read_keyword ?history () = object(self)
   inherit ['a read_keyword_result] engine ?history ()
@@ -646,27 +737,28 @@ class ['a] read_keyword ?history () = object(self)
     let input = Zed_rope.to_string (Zed_edit.text self#edit) in
     try Rk_value(List.assoc input self#keywords) with Not_found -> Rk_error input
 
-  method completion =
+  method! completion =
     let word = Zed_rope.to_string self#input_prev in
-    let keywords = List.filter (fun (keyword, value) -> Zed_utf8.starts_with keyword word) self#keywords in
-    self#set_completion 0 (List.map (fun (keyword, value) -> (keyword, "")) keywords)
+    let keywords = List.filter (fun (keyword, _value) -> Zed_string.starts_with ~prefix:word keyword) self#keywords in
+    self#set_completion 0 (List.map (fun (keyword, _value) -> (keyword, Zed_string.empty ())) keywords)
 end
 
 (* +-----------------------------------------------------------------+
    | Running in a terminal                                           |
    +-----------------------------------------------------------------+ *)
 
-let newline = UChar.of_char '\n'
+let newline_uChar = UChar.of_char '\n'
+let newline = Zed_char.unsafe_of_uChar @@ newline_uChar
 let vline = LTerm_draw.({ top = Light; bottom = Light; left = Blank; right = Blank })
 let reverse_style = { LTerm_style.none with LTerm_style.reverse = Some true }
-let default_prompt = LTerm_text.of_string "# "
+let default_prompt = LTerm_text.of_utf8 "# "
 
 let rec drop count l =
   if count <= 0 then
     l
   else match l with
     | [] -> []
-    | e :: l -> drop (count - 1) l
+    | _ :: l -> drop (count - 1) l
 
 (* Computes the position of the cursor after printing the given styled
    string:
@@ -680,13 +772,15 @@ let rec compute_position cols pos text start stop =
   if start = stop then
     pos
   else
-    let ch, style = text.(start) in
+    let ch, _style = text.(start) in
     if ch = newline then
       compute_position cols { row = pos.row + 1; col = 0 } text (start + 1) stop
-    else if pos.col = cols then
-      compute_position cols { row = pos.row + 1; col = 1 } text (start + 1) stop
     else
-      compute_position cols { pos with col = pos.col + 1 } text (start + 1) stop
+      let width= Zed_char.width ch in
+      if pos.col + width > cols then
+        compute_position cols { row = pos.row + 1; col = width } text (start + 1) stop
+      else
+        compute_position cols { pos with col = pos.col + max 0 width } text (start + 1) stop
 
 (* Return the "real" position of the cursor, i.e. on the screen. *)
 let real_pos cols pos =
@@ -699,13 +793,24 @@ let rec get_index_of_last_displayed_word column columns index words =
   match words with
     | [] ->
         index - 1
-    | (word, suffix) :: words ->
-        let column = column + Zed_utf8.length word in
+    | (word, _suffix) :: words ->
+        let column = column + Zed_string.length word in
         if column <= columns - 1 then
           get_index_of_last_displayed_word (column + 1) columns (index + 1) words
         else
           index - 1
 
+(*let rec get_index_of_last_displayed_word_by_width column columns index words =
+  match words with
+    | [] ->
+        index - 1
+    | (word, _suffix) :: words ->
+        let column = column + Zed_string.(aval_width (width word)) in
+        if column <= columns - 1 then
+          get_index_of_last_displayed_word_by_width (column + 1) columns (index + 1) words
+        else
+          index - 1*)
+
 let draw_styled ctx row col str =
   let size = LTerm_draw.size ctx in
   let rec loop row col idx =
@@ -714,40 +819,35 @@ let draw_styled ctx row col str =
       if ch = newline then
         loop (row + 1) 0 (idx + 1)
       else begin
-        let point = LTerm_draw.point ctx row col in
-        point.LTerm_draw.char <- ch;
-        LTerm_draw.set_style point style;
-        let col = col + 1 in
-        if col = size.cols then
-          loop (row + 1) 0 (idx + 1)
+        let width= max 1 (Zed_char.width ch) in
+        if col + width > size.cols then
+          loop (row + 1) 0 idx
         else
-          loop row col (idx + 1)
+          begin
+            LTerm_draw.draw_char ctx row col ~style ch;
+            loop row (col+width) (idx + 1)
+          end
       end
     end
   in
   loop row col 0
 
-let unsafe_get matrix row col =
-  Array.unsafe_get (Array.unsafe_get matrix row) col
-
 let draw_styled_with_newlines matrix cols row col str =
   let rec loop row col idx =
     if idx < Array.length str then begin
       let ch, style = Array.unsafe_get str idx in
       if ch = newline then begin
-        (unsafe_get matrix row col).LTerm_draw.char <- newline;
+        LTerm_draw.draw_char_matrix matrix row col newline;
         loop (row + 1) 0 (idx + 1)
       end else begin
-        let row, col =
-          if col = cols then
-            (row + 1, 0)
-          else
-            (row, col)
-        in
-        let point = unsafe_get matrix row col in
-        point.LTerm_draw.char <- ch;
-        LTerm_draw.set_style point style;
-        loop row (col + 1) (idx + 1)
+        let width= max 1 (Zed_char.width ch) in
+        if col + width > cols then
+          loop (row + 1) 0 idx
+        else
+          begin
+            LTerm_draw.draw_char_matrix matrix row col ~style ch;
+            loop row (col + width) (idx + 1)
+          end
       end
     end
   in
@@ -892,7 +992,7 @@ object(self)
                   col2 = size.cols;
                 } LTerm_draw.Light;
                 for row = prompt_input_height to total_height - 1 do
-                  (unsafe_get matrix row size.cols).LTerm_draw.char <- newline
+                  LTerm_draw.draw_char_matrix matrix row size.cols newline;
                 done;
 
                 (* Draw the message. *)
@@ -941,7 +1041,7 @@ object(self)
                   col2 = size.cols;
                 } LTerm_draw.Light;
                 for row = prompt_input_height to total_height - 1 do
-                  (unsafe_get matrix row size.cols).LTerm_draw.char <- newline
+                  LTerm_draw.draw_char_matrix matrix row size.cols newline;
                 done;
 
                 (* Draw the completion. *)
@@ -955,8 +1055,8 @@ object(self)
                 let rec loop idx col = function
                   | [] ->
                       ()
-                  | (word, suffix) :: words ->
-                      let len = Zed_utf8.length word in
+                  | (word, _suffix) :: words ->
+                      let len = Zed_string.length word in
                       LTerm_draw.draw_string ctx 0 col word;
                       (* Apply the reverse style if this is the selected word. *)
                       if idx = comp_index then
@@ -1006,23 +1106,15 @@ object(self)
   method draw_success =
     let size = S.value size in
     if size.rows > 0 && size.cols > 0 then begin
-      let styled, position = self#stylise true in
+      let styled, _position = self#stylise true in
       let prompt = S.value prompt in
-      let pos_after_prompt = compute_position size.cols { row = 0; col = 0 } prompt 0 (Array.length prompt) in
-      let pos_after_before = compute_position size.cols pos_after_prompt styled 0 position in
-      let pos_after_styled = compute_position size.cols pos_after_before styled position (Array.length styled) in
-      let total_height = pos_after_styled.row + 1 in
-      let matrix_size = { cols = size.cols + 1; rows = if displayed then max total_height height else total_height } in
-      let matrix = LTerm_draw.make_matrix matrix_size in
-      draw_styled_with_newlines matrix size.cols 0 0 prompt;
-      draw_styled_with_newlines matrix size.cols pos_after_prompt.row pos_after_prompt.col styled;
-      draw_styled_with_newlines matrix size.cols pos_after_styled.row pos_after_styled.col styled_newline;
-      (if displayed then LTerm.move term (-cursor.row) (-cursor.col) else return ()) >>= fun () ->
-      LTerm.print_box_with_newlines term matrix >>= fun () ->
-      LTerm.move term (total_height - Array.length matrix) 0 >>= fun () ->
-      (* Print a newline instead of a movement to ensure scrolling when
-         at the end of screen. *)
-      LTerm.fprint term "\n"
+      (if displayed then
+         LTerm.move term (-cursor.row) (-cursor.col) >>= fun () ->
+         LTerm.clear_screen_next term
+       else
+         return ()) >>= fun () ->
+      LTerm.fprints term prompt >>= fun () ->
+      LTerm.fprintls term styled
     end else
       return ()
 
@@ -1039,7 +1131,7 @@ object(self)
           let matrix_size = { cols = size.cols + 1; rows = height } in
           let matrix = LTerm_draw.make_matrix matrix_size in
           for row = 0 to height - 1 do
-            (unsafe_get matrix row 0).LTerm_draw.char <- newline
+            LTerm_draw.draw_char_matrix matrix row 0 newline;
           done;
           LTerm.move term (-cursor.row) (-cursor.col) >>= fun () ->
           LTerm.print_box_with_newlines term matrix >>= fun () ->
@@ -1100,11 +1192,11 @@ object(self)
                 if resolver = None then
                   match key with
                     | { control = false; meta = false; shift = false; code = Char ch } ->
-                        Zed_macro.add self#macro (Edit (LTerm_edit.Zed (Zed_edit.Insert ch)));
+                        Zed_macro.add self#macro (Edit (LTerm_edit.Zed (Zed_edit.Insert (Zed_char.unsafe_of_uChar ch))));
                         self#insert ch
-                    | { code = Char ch } when LTerm.windows term && UChar.code ch >= 32 ->
+                    | { code = Char ch; _ } when LTerm.windows term && UChar.code ch >= 32 ->
                         (* Windows reports Shift+A for A, ... *)
-                        Zed_macro.add self#macro (Edit (LTerm_edit.Zed (Zed_edit.Insert ch)));
+                        Zed_macro.add self#macro (Edit (LTerm_edit.Zed (Zed_edit.Insert (Zed_char.unsafe_of_uChar ch))));
                         self#insert ch
                     | _ ->
                         ()
@@ -1177,7 +1269,7 @@ object(self)
         end >>= fun () ->
         let temp_fn = self#create_temporary_file_for_external_editor in
         let input = Zed_rope.to_string (Zed_edit.text self#edit) in
-        Lwt_io.with_file ~mode:Output temp_fn (fun oc -> Lwt_io.write_line oc input)
+        Lwt_io.with_file ~mode:Output temp_fn (fun oc -> Lwt_io.write_line oc (Zed_string.to_utf8 input))
         >>= fun () ->
         let editor = self#external_editor in
         Printf.ksprintf Lwt_unix.system "%s %s" editor (Filename.quote temp_fn)
@@ -1195,7 +1287,7 @@ object(self)
                 let s = Zed_utf8.rstrip s in
                 Zed_edit.goto_bot self#context;
                 Zed_edit.replace self#context (Zed_rope.length (Zed_edit.text self#edit))
-                  (Zed_rope.of_string s);
+                  (Zed_rope.of_string (Zed_string.unsafe_of_utf8 s));
                 Lwt.return ())
              (function
                | Unix.Unix_error (err, _, _) ->
